package de.ids_mannheim.korap.openthesaurus;

import java.io.*;
import java.net.URL;
import java.net.URLConnection;
import java.net.HttpURLConnection;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

public class OpenThesaurusData {

    /**
     * Gets the filesize from a URL of a file to be downloaded.
     * This is necessary to know in cases, where a download is resumed
     * (e.g. after the connection is lost), because we don't want to start
     * downloading from the beginning.
     * @param  url URL of a file to download
     * @return filesize in bytes
     */
    private static int getFileSize(URL url) {
        URLConnection conn = null;
        try {
            conn = url.openConnection();
            if(conn instanceof HttpURLConnection) {
                ((HttpURLConnection) conn).setRequestMethod("HEAD");
            }
            conn.getInputStream();
            return conn.getContentLength();
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            if(conn instanceof HttpURLConnection) {
                ((HttpURLConnection)conn).disconnect();
            }
        }
    }

    /**
     * A convenience function that combines downloading and extracting a .zip file
     * (e.g. OpenThesaurus data is provided as a .zip-file) at the same time
     * @param URLString link to the data to download
     * @param thesaurusFilepath filepath to save the extracted data
     * @throws IOException
     */
    public static void getFile(String URLString, String thesaurusFilepath) throws IOException {

        URL url = new URL(URLString);
        int fileSizeToDownload = getFileSize(url);

        File thesaurusFile = new File(thesaurusFilepath);

        // This creates directories in the filepath if they don't yet exist
        String[] thesaurusFilepathSplit = thesaurusFilepath.split(File.separator);
        StringBuilder dirsBeforeFile = new StringBuilder();
        for (int i = 0; i < thesaurusFilepathSplit.length - 1; i++) {
            dirsBeforeFile.append(thesaurusFilepathSplit[i] + File.separator);
        }
        File dirsBeforeFileFile = new File(dirsBeforeFile.toString());
        dirsBeforeFileFile.mkdirs();

        long downloadedSoFar = 0;
        URLConnection urlconnection = url.openConnection();
        urlconnection.setConnectTimeout(10000);
        urlconnection.setReadTimeout(10000);
        FileOutputStream FOS;
        // Resume download if the file already exists, and it's not yet complete
        if (thesaurusFile.exists()) {
            downloadedSoFar = thesaurusFile.length();
            if (downloadedSoFar - fileSizeToDownload >=0)
                return;
            urlconnection.setRequestProperty("Range", "bytes=" + downloadedSoFar + "-");
            FOS = new FileOutputStream(thesaurusFile, true);
        } 
        else {
            FOS = new FileOutputStream(thesaurusFile, false);
        }

        BufferedOutputStream outputStream = new BufferedOutputStream(FOS);
        urlconnection.connect();
        ZipInputStream zipIn = new ZipInputStream(urlconnection.getInputStream());
        ZipEntry entry = zipIn.getNextEntry();
        while (entry != null) {
            if (!entry.isDirectory()) {
                // if the entry is a file, extracts it
                extractFile(zipIn, outputStream);
            }
            zipIn.closeEntry();
            entry = zipIn.getNextEntry();
        }
        zipIn.close();
        outputStream.flush();
        outputStream.close();
    }

    /**
     * Extracts the data from the ZipInputStream in 8192 byte chunks to a given
     * OutputStream
     * @param zipIn ZipInputStream with the data to extract
     * @param bos OutputStream to extract the data to
     * @throws IOException
     */
    private static void extractFile(ZipInputStream zipIn, OutputStream bos) throws IOException {
        byte[] bytesIn = new byte[8192];
        int read = 0;
        while ((read = zipIn.read(bytesIn)) != -1) {
            bos.write(bytesIn, 0, read);
        }
    }
    /* // Test functionality:
    public static void main(String[] args) throws IOException {
        String url = "https://www.openthesaurus.de/export/OpenThesaurus-Textversion.zip";
        String thesaurusFilepath = "openthesaurus_data/openthesaurus.txt";

        getFile(url, thesaurusFilepath);
    }
     */
}
